#!/usr/bin/python3
# -*- coding: UTF-8 -*-

from ..util.tools import Tools

import os, subprocess, socket, re, queue, threading, sys, time, json, hashlib

from ..config import Configuration
from ..util.logger import Logger
from ..util.getter import Getter

class PathGetter:

    words = []
    q = queue.Queue()
    added = []
    last_start = []
    ignore_until = ''
    current_getter = None
    current_uri = ''
    running=True
    duplicated=0
    last_item=""
    paused=True
    skip_current=False

    def __init__(self):
        pass

    def load_iplist(self):

        if Configuration.threads_data is not None:
            try:
                for i in Configuration.threads_data:
                    self.last_start.append(Configuration.threads_data[i])
            except:
                pass

        if Configuration.nmap_file != '':
            import xml.etree.ElementTree as xml
            tree = xml.parse(Configuration.nmap_file)
            root = tree.getroot()
            for h in root.iter('host'):
                dict_item = {
                    'valid': False,
                    'ports': []
                }
                if h.tag == 'host':
                    for c in h:
                        if c.tag == 'address':
                            if c.attrib['addr'] and c.attrib['addrtype'] == 'ipv4':
                                dict_item['ip'] = c.attrib['addr']

                        elif c.tag == 'hostnames':
                            for names in list(c):
                                if names.attrib['name']:
                                    dict_item['hostname'] = names.attrib['name']

                        elif c.tag == 'ports':
                            for port in list(c):
                                if port.tag == 'port' and port.attrib.get('protocol', '').lower() == "tcp":
                                    pi = int(port.attrib['portid'])
                                    for p in list(port):
                                        if p.tag == 'state' and p.attrib.get('state', '').lower() == "open":
                                            if pi not in dict_item['ports']:
                                                dict_item['ports'].append(pi)
                                        elif p.tag == 'service':
                                            if p.attrib['name'] == 'https' or p.attrib['name'] == 'http':
                                                dict_item['valid'] = True
                                            else:
                                                servicefp = p.attrib.get('servicefp', '')
                                                if 'HTTP/' in servicefp or 'SSL' in servicefp:
                                                    dict_item['valid'] = True

                if dict_item.get('valid', False) is False:
                    continue

                if dict_item.get('ip', '') == '' and dict_item.get('hostname', '') != '':
                    dict_item['ip'] = dict_item['hostname']

                if dict_item.get('ip', '') == '':
                    continue

                ip = dict_item['ip']
                for pi in dict_item['ports']:
                    line = f'{ip}:{pi}'

                    if self.ignore_until == '' and line in self.last_start:
                        self.ignore_until = line

                    if line not in Configuration.ipaddresses:
                        Configuration.ipaddresses.append(line.strip('\r\n\t').strip())
                        self.last_item = line

        else:
            with open(Configuration.ip_list, 'r', encoding="ascii", errors="surrogateescape") as f:
                line = f.readline()
                while line:
                    if line.endswith('\n'):
                        line = line[:-1]
                    if line.endswith('\r'):
                        line = line[:-1]

                    line = ''.join(filter(Tools.permited_char, line)).strip()

                    if self.ignore_until == '' and line in self.last_start:
                        self.ignore_until = line

                    if line not in Configuration.ipaddresses:
                        Configuration.ipaddresses.append(line.strip('\r\n\t').strip())
                        self.last_item = line
                    #else:
                    #    self.duplicated+=1

                    try:
                        line = f.readline()
                    except:
                        pass

    def len(self):
        return len(Configuration.ipaddresses)

    def run(self):

        t = threading.Thread(target=self.worker)
        t.daemon = True
        t.start()

        t_status = threading.Thread(target=self.status_worker)
        t_status.daemon = True
        t_status.start()

        count = 0
        with self.q.mutex:
            self.q.queue.clear()

        count += 1
        self.q.put(Configuration.target)

        #if len(list(self.q.queue)) > 0:
        if count > 0:
            self.paused = False

            self.q.join()  # block until all tasks are done
            sys.stdout.write("\033[K")  # Clear to the end of line

    def worker(self):
        try:
            while self.running:

                while self.paused:
                    time.sleep(0.3)

                item = self.q.get()

                self.current_uri = item
                self.current_getter = Getter(Configuration.ipaddresses, False)
                self.current_getter.ignore_until = self.ignore_until
                self.current_getter.run(item)

                if Configuration.verbose > 3:
                    Logger.pl('\n{*} {W}Finishing %s{W}' % item)

                self.ignore_until = ''
                self.q.task_done()
        except KeyboardInterrupt as e:
            raise e

    def testing_base(self):
        return self.current_uri == Configuration.target

    def pause(self):
        self.paused = True
        self.running = False
        self.save_status()
        self.current_getter.pause()

    def skip(self):
        self.ignore_until = ''
        self.save_status(True)
        self.running = False
        self.paused = False
        self.current_getter.stop()
        with self.q.mutex:
            self.q.queue.clear()

    def save_status(self, skip_current=False):
        paths_found = self.current_getter.path_found

        for u in paths_found:
            if u.endswith('/'):
                u = u[:-1]
            if u not in self.added:
                self.added.append(u)
                self.q.put(u)

        dt = {
            "command": Configuration.cmd_line,
            "current_path": self.current_uri,
            "skip_current": skip_current,
            "paths": self.added,
            "deep_links": Getter.deep_links,
            "threads": self.current_getter.last
         }

        with open("webfinder.restore", "w") as text_file:
            text_file.write(json.dumps(dt))

    def status_worker(self):
        try:
            while self.running:
                try:
                    if self.current_getter is None:
                        time.sleep(1)
                        continue

                    if self.paused:
                        time.sleep(1)
                        continue

                    self.save_status()

                except:
                    raise
                time.sleep(10)
        except KeyboardInterrupt:
            pass
            




#!/usr/bin/python3
# -*- coding: UTF-8 -*-
from requests import Response
from webfinder.util.tools import Tools


class ResultPattern(object):
    RES_LIMIT = 0.07
    ENTROPY_LIMIT = 0.10
    _status_code = 0
    _length = -1
    _min_length = 0
    _max_length = 0
    _entropy = 0
    _min_entropy = 0
    _max_entropy = 0
    _content_type = None

    def __init__(self,
                 status_code: int = 200,
                 length: int = -1,
                 entropy: float = -1.0,
                 content_type: str = None):
        self._status_code = status_code
        if length > 0:
            self._length = length
            self._min_length = float(self._length) * (1.0 - ResultPattern.RES_LIMIT)
            self._max_length = float(self._length) * (1.0 + ResultPattern.RES_LIMIT)
        if entropy > 0:
            self._entropy = entropy
            self._min_entropy = float(self._entropy) * (1.0 - ResultPattern.ENTROPY_LIMIT)
            self._max_entropy = float(self._entropy) * (1.0 + ResultPattern.ENTROPY_LIMIT)

        if isinstance(content_type, str):
            self._content_type = content_type.lower()

    def is_valid_result(self,
                        status_code: int = 200,
                        length: int = -1,
                        entropy: float = -1.0,
                        content_type: str = None) -> bool:
        from webfinder.config import Configuration

        if status_code != self._status_code:
            return False

        if self._length == -1 or length == -1:
            return True

        if self._min_length <= length <= self._max_length:

            ct = False
            ent = False

            if self._content_type is None or content_type is None or not isinstance(content_type, str):
                ct = True

            if Configuration.no_content_type:
                ct = True

            if ct is False and self._content_type == content_type.lower():
                ct = True

            if self._entropy == -1 or entropy == -1:
                ent = True

            if Configuration.no_entropy:
                ent = True

            if ent is False and self._min_entropy <= entropy <= self._max_entropy:
                ent = True

            if ct is True and ent is True:
                return True

        return False

    def is_valid_response(self, response) -> bool:
        if not isinstance(response, Response):
            return False

        txt = response.content
        entropy = Tools.get_entropy(txt)
        ct = next(iter([
            str(f"{v};").lower().split(';')[0].strip()
            for k, v in response.headers.items()
            if k is not None and k.strip != ''
            and k.lower() == 'content-type'
        ]), None)

        return self.is_valid_result(
            status_code=response.status_code,
            length=len(txt),
            entropy=entropy,
            content_type=ct
        )

    @staticmethod
    def from_response(response):
        if not isinstance(response, Response):
            raise Exception('Invalid response!')

        txt = response.content
        entropy = Tools.get_entropy(txt)
        ct = next(iter([
            str(f"{v};").lower().split(';')[0].strip()
            for k, v in response.headers.items()
            if k is not None and k.strip != ''
            and k.lower() == 'content-type'
        ]), None)

        return ResultPattern(
            status_code=response.status_code,
            length=len(txt),
            entropy=entropy,
            content_type=ct
        )




# Source
# https://yzlin.medium.com/python-requests-ssl-ip-binding-6df25a9a8f6a

import ssl

import requests
from requests.adapters import HTTPAdapter


class HostSSLContext(ssl.SSLContext):
    """
    A SSL Context for wrapping socket with specific hostname for requesting certificate.
    """

    def __new__(cls, hostname):
        self = super(HostSSLContext, cls).__new__(cls, protocol=ssl.PROTOCOL_TLS_CLIENT)
        self.check_hostname = False
        self.options |= 0x4  # => Enable UNSAFE LEGACY RENEGOTIATION
        return self

    def __init__(self, hostname):
        super(HostSSLContext, self).__init__()
        self._hostname = hostname
        self.check_hostname = False
        self.options |= 0x4  # => Enable UNSAFE LEGACY RENEGOTIATION

        # Disable SSLv2 & SSLv3
        # Refer to https://docs.python.org/2/library/ssl.html#security-considerations
        # for more details.
        #self.options |= ssl.OP_NO_SSLv2
        #self.options |= ssl.OP_NO_SSLv3

    def change_server_hostname(self, hostname):
        self._hostname = hostname

    def wrap_socket(self, *args, **kwargs):
        kwargs['server_hostname'] = self._hostname
        return super(HostSSLContext, self).wrap_socket(*args, **kwargs)


class HostHeaderSSLAdapter(HTTPAdapter):
    """
    A HTTPS Adapter for Python Requests that sets the hostname for certificate
    verification based on the Host header.
    This allows requesting the IP address directly via HTTPS without getting
    a "hostname doesn't match" exception.
    Example usage:
        >>> s.mount('https://', HostHeaderSSLAdapter())
        >>> s.get("https://93.184.216.34", headers={"Host": "example.org"})
    """

    _hostname = None

    def send(self, request, **kwargs):
        # HTTP headers are case-insensitive (RFC 7230)
        hostname = None
        for header in request.headers:
            if header.lower() == 'host':
                hostname = request.headers[header]
                break

        if hostname and self._hostname != hostname:
            self._hostname = hostname
            context = HostSSLContext(hostname)
            self.init_poolmanager(self._pool_connections, self._pool_maxsize, block=self._pool_block,
                                  ssl_context=context)

        connection_pool_kwargs = self.poolmanager.connection_pool_kw

        if hostname:
            connection_pool_kwargs['assert_hostname'] = hostname
        elif 'assert_hostname' in connection_pool_kwargs:
            # cleanup previous SSL host binding
            connection_pool_kwargs.pop('assert_hostname', None)
            if self._hostname:
                self._hostname = None
                self.init_poolmanager(self._pool_connections, self._pool_maxsize, block=self._pool_block)

        return super(HostHeaderSSLAdapter, self).send(request, **kwargs)